import os

_TEMPLATES = os.path.join(os.path.split(os.path.abspath(__file__))[0], "templates")
_PRO_BLANK_TEMPLATE = os.path.join(_TEMPLATES, "pro", "blank_pro_project", "blank_pro_project.aprx")
_PRO_BLANK_LAYOUT = os.path.join(_TEMPLATES, "pro", "blank_layout.pagx")

_BLANK_FEATURE_LAYER = os.path.join(_TEMPLATES, "pro", "blank_layer.lyr")
_BLANK_RASTER_LAYER = os.path.join(_TEMPLATES, "pro", "blank_raster_layer.lyr")


import os
import logging
log = logging.getLogger("amaptor")

import arcpy

from amaptor.version_check import PRO, ARCMAP, mapping, mp
from amaptor.classes.map import Map
from amaptor.classes.layout import Layout
from amaptor.classes.map_frame import MapFrame

from amaptor.constants import _TEMPLATES, _PRO_BLANK_LAYOUT
from amaptor.functions import _import_mxd_to_new_pro_project

from amaptor.errors import *


class Project(object):
	"""
		An ArcGIS Pro Project or an ArcMap map document - maps in ArcGIS Pro and data frames in ArcMap are Map class attached to this project
		Access to the underlying object is provided using name ArcGISProProject and ArcMapDocument
	"""

	def __init__(self, path):

		self.maps = []  # stores list of included maps/dataframes
		self.layouts = []
		self.path = None  # will be set after any conversion to current version of ArcGIS is done (aprx->mxd or vice versa)
		self.map_document = None
		self.arcgis_pro_project = None

		self.primary_document = None  # will be an alias for either self.map_document or self.arcgis_pro_project depending on what we're working with - makes it easier for items where API isn't different

		# this conditional tree is getting a little beefy now - could probably be refactored
		if PRO:
			if path == "CURRENT":
				self.path = "CURRENT"  # will be redirected to actual path after setup
			elif path.endswith("aprx"):
				self.path = path
			elif path.endswith("mxd"):
				self.path = _import_mxd_to_new_pro_project(path)
			else:
				raise ValueError("Project or MXD path not recognized as an ArcGIS compatible file (.aprx or .mxd)")

			self._pro_setup()

		else:  # ArcMap

			if path == "CURRENT":
				self.path = "CURRENT"
				self._arcmap_setup()
			elif path.endswith("mxd"):
				self.path = path  # we'll overwrite once set up for "CURRENT"
				self._arcmap_setup()
			elif path.endswith("aprx"):
				# I need to find a way to create blank ArcGIS Pro projects here - may need to include one as a template to copy, but that seems silly/buggy.
				# planned approach is to create a Pro project in a temporary location, and import the map document provided.
				raise MapNotImplementedError("Support for Pro Projects in ArcMap is not possible. Please provide an MXD template to work with.")
			else:
				raise ValueError("Project or MXD path not recognized as an ArcGIS compatible file (.aprx or .mxd)")

		if path == "CURRENT":
			self.path = self.primary_document.filePath

	def _pro_setup(self):
		"""
			Sets up the data based on the ArcGIS Pro Project. Only called if working with arcpy.mp and after any needed
			conversion from Map Document to Pro Project is done.
		:return: None
		"""
		self.arcgis_pro_project = mp.ArcGISProject(self.path)
		self.primary_document = self.arcgis_pro_project
		for l_map in self.arcgis_pro_project.listMaps():
			self.maps.append(Map(self, l_map))

		for layout in self.primary_document.listLayouts():
			self.layouts.append(Layout(layout, self))

		for map in self.maps:
			map._index_frames()

	def _arcmap_setup(self):
		"""
			Sets up data based on an ArcGIS Map Document. Only called if working with arcpy.mapping and after any
			needed conversion from Pro Project to map docusment is done (can we go that way?)
		:return: None
		"""
		self.map_document = mapping.MapDocument(self.path)
		self.primary_document = self.map_document
		for l_map in mapping.ListDataFrames(self.map_document):
			self.maps.append(Map(self, l_map))

	def list_maps(self):
		"""
			Provided to give a similar interface to ArcGIS Pro - Project.maps is also publically accessible
		:return:
		"""
		return self.maps

	def find_layer(self, path, find_all=True):
		"""
			Finds a layer in all maps by searching for the path. By default finds all, but can find just the first one too
		:param path: the full path of the data source for the layer
		:param find_all: When True, reutrns a list of amaptor.Map instances that match. When false, returns only the first match
		:return: list of amaptor.map instances or a single amaptor.map instance.
		"""
		layers = []
		for map in self.maps:
			try:
				new_layers = map.find_layer(path=path, find_all=find_all)
			except LayerNotFoundError:
				continue

			if find_all:  # if it didn't find any, we would have raised an exception, so we have something
				layers += new_layers
			else:  # if we were only supposed to get one, return it
				return new_layers

		if len(layers) == 0:
			raise LayerNotFoundError()

		return layers

	@property
	def active_map(self):
		"""
			Returns the active map object or data frame as determined by get_active_map()
		:return:
		"""
		return self.get_active_map()

	@property
	def map_names(self):
		"""
			A convenience function to get a list of map names
		:return:
		"""
		return [l_map.name for l_map in self.maps]

	@property
	def default_geodatabase(self):
		"""
			Returns the Project's default geodatabase in Pro, and the current workspace (arcpy.env.workspace) in ArcMap.
			If arcpy.env.workspace is None, creates a GDB in same folder as map document and returns that value, to ensure
			that this function always returns a usable workspace. If a GDB is created, this function does NOT set arcpy.env.workspace
			to that, so as not to interfere with other operations. Do that explicitly if that behavior is desired.
		:return:
		"""
		if PRO:
			return self.arcgis_pro_project.defaultGeodatabase
		else:
			if arcpy.env.workspace is not None:
				return arcpy.env.workspace
			else:
				folder_path = os.path.split(self.path)[0]
				name = "amaptor_default_gdb"
				arcpy.CreateFileGDB_management(folder_path, name)
				return os.path.join(folder_path, name)

	@default_geodatabase.setter
	def default_geodatabase(self, value):
		"""
			Sets the default geodatabase in Pro and sets arcpy.env.workspace in ArcMap
		:param value:
		:return:
		"""
		if PRO:
			self.arcgis_pro_project.defaultGeodatabase = value
		else:
			arcpy.env.workspace = value

	def find_map(self, name):
		"""
			Given a map name, returns the map object or raises MapNotFoundError
		:param name: name of map to find.
		:return: amaptor.Map instance
		"""
		for l_map in self.maps:
			if l_map.name == name:
				return l_map
		else:
			raise MapNotFoundError(name)

	def check_map_name(self, name):
		"""
			Checks to see if the project or map document already has a map or data frame with a given name.
			Since names must be unique in ArcGIS Pro, this code helps check before adding new maps
		:param name: name of map to check for
		:return: None. Raises an error if name is taken
		"""

		try:
			self.find_map(name)  # it finds one, then raise a MapExists error
			raise MapExists(name)
		except MapNotFoundError:
			pass  # it's great if it's not found

	def new_map(self, name, template_map=os.path.join(_TEMPLATES, "arcmap", "pro_import_map_template.mxd"), template_df_name="_rename_template_amaptor"):
		"""
			PRO ONLY. Creates a new map in the current project using a hack (importing a blank map document, and renaming data frame)
			Warning: Only works in Pro due to workaround. There isn't a way to add a data frame from arcpy.mapping.
			In the future, this could potentially work in arcmap by transparently working with a separate map document
			in the background (creating a project, map, and layout for those items and linking them into this project).
		:param name: The name to give the imported map
		:param template_map: The map document to import. If we're just going with a blank new map, leave as default. To
							import some other template as your base, provide a path to a document importable to ArcGIS Pro'
							.importDocument function for projects.
		:param template_df_name: The current name of the imported map document for renaming. Only needs to be set if template_map is overridden
		:return: amaptor.Map instance - also added to the map document, but returned for immediate use.
		"""

		if ARCMAP:
			raise MapNotImplementedError("ArcMap doesn't suppport adding data frames to map documents from Python")

		self.check_map_name(name)

		# step 1: import
		self.primary_document.importDocument(template_map, include_layout=False)

		# step 2: set up for amaptor and rename to match passed value
		for l_map in self.primary_document.listMaps():
			if l_map.name == template_df_name:
				l_map.name = name
				new_map = Map(self, l_map)
				self.maps.append(new_map)
				return new_map
		else:  # if it's not found
			raise MapNotFoundError(template_df_name, "Map was inserted, but could not be found after insertion. If you provided a custom" \
								   "template, check that the name you provided for template_df_name matches the name of " \
								   "the data frame you want to use from the map document.")

	def check_layout_name(self, name):
		"""
			PRO ONLY. Given the name of a layout, confirms it doesn't exist and raised amaptor.LayoutExists if it's found
		:param name: the case sensitive name of an existing layout to find
		:return: None. Raises amaptor.LayoutExists if layout with name exists.
		"""
		try:
			self.find_layout(name)  # it finds one, then raise a MapExists error
			raise LayoutExists(name)
		except LayoutNotFoundError:
			pass  # it's great if it's not found

	def find_layout(self, name):
		"""
			PRO ONLY. Given a layout name, returns the amaptor.Layout object or raises LayoutNotFoundError
		:param name: the name of the layout to find.
		:return: amaptor.Layout instance with given name.
		"""
		for layout in self.layouts:
			if layout.name == name:
				return layout
		else:
			raise LayoutNotFoundError(name)

	def new_layout(self, name, template_layout=_PRO_BLANK_LAYOUT, template_name="_pro_blank_layout_template"):
		"""
			PRO ONLY. Adds a new layout to an ArcGIS Pro Project by importing a saved blank layout. Alternatively,
			 you can provide an importable layout document (.pagx) for ArcGIS Pro, and then provide that layout's name
			 as template_name so that it can be renamed, and the provided template will be used instead of a blank.
		:param name: The name to give the new layout
		:param template_layout: The template to use for creating the layout (an ArcGIS Pro .pagx file).
			If none is provided, uses a blank template
		:param template_name: The name of the layout in the template. Only define this value if you also provide a new
			template layout and the name should match the layout name in the template. This parameter is used to find
			the inserted template and rename it. Strange things will happen if this value does not match the name of the
			layout in the template_layout.
		:return: amaptor.Layout instance. This layout will already have been added to the project, but is returned for
			convenience.
		"""
		if ARCMAP:
			raise MapNotImplementedError("ArcMap doesn't suppport adding data frames to map documents from Python")

		# step 1: import
		self.primary_document.importDocument(template_layout)

		# step 2: set up for amaptor and rename to match passed value
		for layout in self.primary_document.listLayouts():
			if layout.name == template_name:
				layout.name = name
				new_layout = Layout(layout, self)
				self.layouts.append(new_layout)
				return new_layout
		else:
			raise LayoutNotFoundError("Layout was inserted, but could not be found after insertion. If you provided a custom" \
								   "template, check that the name you provided for template_name matches the name of " \
								   "the original mxd file from which you derived the layout you're importing")

	def get_active_map(self, use_pro_backup=True):
		"""
			Emulates functionality of arcpy.mapping(mxd).activeDataFrame. In ArcMap, it returns the amaptor.Map object
			that corresponds to that active Data Frame. In Pro, which doesn't have the concept of active maps, it by
			default returns the first map in the document. If use_pro_backup is set to False, it will instead
			raise amaptor.MapNotImplementedError
		:param use_pro_backup: When True, it uses the first map in the ArcGIS Pro project, since Pro doesn't have a way
		 to get the active map.
		:return: amaptor.Map instance
		"""
		if ARCMAP:
			for each_map in self.maps:
				if each_map.name == self.primary_document.activeDataFrame.name:
					return each_map
		else:
			# ArcGIS Pro doesn't have this capability, so we'll just use the first map (unfortunately).
			if use_pro_backup:
				return self.maps[0]
			else:
				raise MapNotImplementedError("ArcGIS Pro does not provide an interface to the active map")

	def save(self):
		"""
			Saves the project or map document in place.
		:return: None
		"""
		self.primary_document.save()

	def save_a_copy(self, path):
		"""
			Saves the project or map document to the provided path.
		:param path: the new path to save the copy of the document to.
		:return: None
		"""
		self.primary_document.saveACopy(path)

	def to_package(self, output_file, summary, tags, **kwargs):
		"""
			Though it's not normally a mapping method, packaging concepts need translation between the two versions, so
			we've included to_package for maps and projects. In ArcGIS Pro, project.to_package will create a Project Package
			and map.to_package will create a map package. In ArcMap, both will create a map package. Extra **kwargs beyond
			output path are only passed through to Pro Package command, not to map packages. To pass kwargs through to a map
			package, use a map object's to_package method.
		:param output_file: the path to output the package to
		:param kwargs: dictionary of kwargs to pass through to project packaging in Pro.
		:return: None
		"""

		log.warning("Warning: Saving project to export package")
		self.save()

		if PRO:
			arcpy.PackageProject_management(self.path, output_file, summary=summary, tags=tags, **kwargs)
		else:
			arcpy.PackageMap_management(self.path, output_file, summary=summary, tags=tags)

	def replace_text(self, text, replacement):
		"""
			Given a string and a replacement value, finds all instances (in the current map document or project)
			of that text in text elements and titles, and replaces those instances with the new value.
			Useful for creating your own variables like {species} or {field_id} in map templates.
			Careful when using this - in Pro, it will search all Layouts and replace the string.
			If you are concerned and want single layout behavior, use the same function on the Layout class.
		:param text:
		:param replacement:
		:return:
		"""

		if ARCMAP:
			for elm in arcpy.mapping.ListLayoutElements(self.primary_document, "TEXT_ELEMENT"):
				elm.text = elm.text.replace(text, replacement)
		else:
			for layout in self.layouts:  # in pro, iterate through Layout objects instead and replace in all
				layout.replace_text(text, replacement)



import logging
log = logging.getLogger("amaptor")

import arcpy

from amaptor.classes.map_frame import MapFrame
from amaptor.version_check import PRO
from amaptor.errors import MapFrameNotFoundError, ElementNotFoundError, NotSupportedError

class Layout(object):
	"""
		Replicates Layouts so that we can do some nice things behind the scenes.

		In ArcMap, a single layout is created - that way, a layout can safely be retrieved for all documents and modifying
		properties of a layout modifies corresponding map document properties.
	"""

	def __init__(self, layout_object, project):
		self._layout_object = layout_object
		self.project = project
		self.frames = [MapFrame(frame, self) for frame in self._layout_object.listElements("MAPFRAME_ELEMENT")]  # frames connect back to maps, this connects to a project
		self.elements = self.list_elements()

	@property
	def name(self):
		"""
			Corresponds to the name of a layout in Pro and the Map Document's "title" property in ArcMap
		:return:
		"""
		if PRO:
			return self._layout_object.name
		else:
			return self.project.primary_document.title

	@name.setter
	def name(self, value):
		if PRO:
			self._layout_object.name = value
		else:
			self.project.primary_document.title = value

	def list_elements(self):
		self.elements = self._layout_object.listElements()
		return self.elements

	def export_to_png(self, out_path, resolution=300):
		"""
			Currently Pro only - needs refactoring to support ArcMap and Pro (should export map document in ArcMap).
			Also needs refactoring to combine Map and Layout export code.
		:param out_path:
		:param resolution:
		:return:
		"""
		self._layout_object.exportToPNG(out_path, resolution)

	def find_map_frame(self, name):
		"""
			Finds the map frame with a given name
		:param name: the name of the frame to find
		:return: MapFrame object
		"""

		if PRO:
			for frame in self.frames:
				if name == frame.name:
					return frame

			raise MapFrameNotFoundError(name=name)
		else:
			raise NotSupportedError("Map Frames are not supported in ArcMap")

	def find_element(self, name):
		"""
			Finds the first element matching the provided name
		:param name:
		:return:
		"""
		if PRO:
			for element in self.elements:
				if element.name == name:
					return element

			raise ElementNotFoundError(name)
		else:
			raise NotSupportedError("Element actions are not supported in ArcMap")

	def toggle_element(self, name_or_element, visibility="TOGGLE"):
		"""
			Given an element name, toggles, makes visible, or makes invisible that element.
		:param name_or_element: a string name of an element, or an element object
		:param visibility: Controls the action. Valid values are boolean (True, False) or the keyword "TOGGLE" which
			switches its current visibility state.
		:return:
		"""
		if PRO:
			if not isinstance(name_or_element, arcpy._mp.GraphicElement):
				element = self.find_element(name_or_element)
			else:
				element = name_or_element

			if visibility is True or visibility is False:
				element.visible = visibility
			elif visibility == "TOGGLE":
				if element.visible is True:
					element.visible = False
				else:
					element.visible = True
			else:
				raise ValueError("parameter visibility must be either a boolean value, (True, False) or the keyword \"TOGGLE\".")

	def export_to_pdf(self, out_path, **kwargs):
		self._layout_object.exportToPDF(out_path, **kwargs)

	def replace_text(self, text, replacement):
		"""
			Single layout analogue of Project.replace_text. Given a string and a replacement value, replaces all
			instances of that string in all text elements in the layout. Useful for having template strings in a map
			document
		:param text:
		:param replacement:
		:return:
		"""

		for elm in self._layout_object.listElements("TEXT_ELEMENT"):
			elm.text = elm.text.replace(text, replacement)



import logging
log = logging.getLogger("amaptor")

from amaptor.version_check import ARCMAP

if ARCMAP:
	FileExistsError = OSError  # define it for Python 2, basically raise an OSError in that case
	FileNotFoundError = OSError

class MapExists(FileExistsError):
	"""
		Raised when a map or data frame already exists because a new one cannot be added with the same name
	"""
	def __init__(self, map_name, **kwargs):
		self.map_name = map_name
		super(MapExists, self).__init__(**kwargs)

	def __repr__(self):
		log.error("Map with name {} already exists.".format(self.map_name))


class LayoutExists(FileExistsError):
	"""
		Raised when a layout (in Pro only) already exists because a new one cannot be added with the same name
	"""
	def __init__(self, layout_name, **kwargs):
		self.layout_name = layout_name
		super(LayoutExists, self).__init__(**kwargs)

	def __repr__(self):
		log.error("Layout with name {} already exists.".format(self.layout_name))

class _AmaptorNotFoundError(FileNotFoundError):
	"""
		Raised when a process is trying to find a specific layout (Pro only) and it is not found.
	"""
	def __init__(self, name, extra_text=None, **kwargs):
		self.item_name = name
		self.extra_text = extra_text
		super(_AmaptorNotFoundError, self).__init__(**kwargs)

	def __repr__(self):
		log.error("{} with name {} does not exist. {}".format(self.type, self.item_name, self.extra_text))

class MapNotFoundError(_AmaptorNotFoundError):
	"""
		Raised when a process is trying to find a specific map and it is not found.
	"""
	def __init__(self, *args, **kwargs):
		self.type = "Map"
		super(MapNotFoundError, self).__init__(*args, **kwargs)

class LayoutNotFoundError(_AmaptorNotFoundError):
	"""
		Raised when a process is trying to find a specific layout (Pro only) and it is not found.
	"""
	def __init__(self, *args, **kwargs):
		self.type = "Layout"
		super(LayoutNotFoundError, self).__init__(*args, **kwargs)

class MapFrameNotFoundError(_AmaptorNotFoundError):
	"""
		Raised when a process is trying to find a specific layout (Pro only) and it is not found.
	"""
	def __init__(self, *args, **kwargs):
		self.type = "MapFrame"
		super(MapFrameNotFoundError, self).__init__(*args, **kwargs)

class ElementNotFoundError(_AmaptorNotFoundError):
	"""
		Raised when searching a layout for an element, and it's not found
	"""
	def __init__(self, *args, **kwargs):
		self.type = "Element"
		super(ElementNotFoundError, self).__init__(*args, **kwargs)

class MapNotImplementedError(NotImplementedError):
	"""
		Raised when a feature is only implemented for either ArcMap or ArcGIS Pro and code tries to use it in the environment
		it is not applicable to.
	"""
	def __init__(self, feature_name, arcmap_or_pro, **kwargs):
		log.error("Feature {} is not implemented for Arc {}.".format(feature_name, arcmap_or_pro))
		super(MapNotImplementedError, self).__init__(**kwargs)
	# for use when a specific mapping function not implemented


class LayerNotFoundError(ValueError):
	"""
		Raised when searching for a specific layer and it isn't found.
	"""
	pass  # for use when looking up layers


class EmptyFieldError(ValueError):
	def __init__(self, field, description, **kwargs):
		self.description = description
		self.field = field
		super(EmptyFieldError, self).__init__(**kwargs)

	def __repr__(self):
		log.error("{} is empty or missing. {}".format(self.field, self.description))

class NotSupportedError(NotImplementedError):
	"""
		Raised when a feature is not supported by ArcGIS itself.
	"""
	def __init__(self, message, **kwargs):
		log.error("Not Supported: {}.".format(message))
		super(NotSupportedError, self).__init__(**kwargs)
	# for use when a specific mapping function not implemented


import logging
log = logging.getLogger("amaptor")

from amaptor.version_check import mp
from amaptor.errors import MapNotFoundError
#from amaptor.classes import map

class MapFrame(object):
	def __init__(self, map_frame_object, layout):
		self._map_frame_object = map_frame_object
		self.layout = layout

		try:
			self._map = layout.project.find_map(map_frame_object.map.name)  # find the map that this relates to
		except MapNotFoundError:
			self._map = None  # this is ok, because if we just created the frame, it may not be findable (as when we import a layout), but it should be set later

	def _set_map(self, amaptor_map):

		#if not isinstance(amaptor_map, map.Map):  # probably getting a circular import here, why it's commented out
		#	raise MapNotFoundError("Provided map is either None or not an instance of amaptor.classes.Map")

		self._map = amaptor_map
		self._map._index_frames()  # have it reindex all of the frames and maps it has
		self._map_frame_object.map = amaptor_map.map_object

	def set_extent(self, extent_object):
		self._map_frame_object.camera.setExtent(extent_object)

	def get_extent(self):
		return self._map_frame_object.camera.getExtent()

	@property
	def name(self):
		return self._map_frame_object.name

	@name.setter
	def name(self, value):
		self._map_frame_object.name = value

	@property
	def map(self):
		return self._map

	@map.setter
	def map(self, value):
		self._set_map(value)




import logging
log = logging.getLogger("amaptor")

# Only one of these will be true, but lets people test against the item of their choice (if amaptor.ARCMAP:, etc)
ARCMAP = None
PRO = None

MAP_EXTENSION = None

try:
	from arcpy import mapping
	ARCMAP = True
	PRO = False
	MAP_EXTENSION = "mxd"
	log.debug("Found ArcGIS Desktop (arcpy.mapping)")
	mp = None  # define so can always be imported
except ImportError:
	try:
		from arcpy import mp
		ARCMAP = False
		PRO = True
		MAP_EXTENSION = "aprx"
		log.debug("Found ArcGIS Pro (arcpy.mp)")
		mapping = None  # define so can always be imported
	except ImportError:
		print("You must run {} on a Python installation that has arcpy installed".format(__name__))
		raise


import os
import tempfile
import warnings

import arcpy

from amaptor.version_check import log, mp, PRO, mapping
from amaptor.errors import LayerNotFoundError, NotSupportedError
from amaptor.constants import _PRO_BLANK_TEMPLATE


def _import_mxd_to_new_pro_project(mxd, blank_pro_template=_PRO_BLANK_TEMPLATE, default_gdb="TEMP"):
	"""
		Handles importing an ArcMap Document into an ArcGIS Pro Project. Default Geodatabase is "TEMP" by default, indicating
		a temporary gdb should be created. It can also be "KEEP" to leave it alone, or it can be a path
	:param mxd:
	:param blank_pro_template:
	:param default_gdb:
	:return:
	"""

	log.warning("WARNING: Importing MXD to new Pro Project - if you call .save() it will not save back to original MXD. Use .save_a_copy('new_path') instead.")

	# can safely assume that if this is called, we're running on Pro and that's already been checked

	# copy blank project to new location - template is 1.3+
	blank_project = mp.ArcGISProject(blank_pro_template)
	new_temp_project = tempfile.mktemp(".aprx", "pro_project_import")
	blank_project.saveACopy(new_temp_project)
	del(blank_project)

	# strictly speaking, we don't need to destroy and recreate - should be able to edit original without saving - doing this just to keep things clear
	project = mp.ArcGISProject(new_temp_project)
	project.importDocument(mxd, include_layout=True)

	if default_gdb != "KEEP":  # if we're supposed to modify it
		if default_gdb == "TEMP":
			new_default_gdb = tempfile.mktemp(prefix="amaptor_default_geodatabase", suffix=".gdb")
			arcpy.CreateFileGDB_management(os.path.split(new_default_gdb)[0], os.path.split(new_default_gdb)[1])
			project.defaultGeodatabase = new_default_gdb
		else:  # if it's not KEEP or TEMP it must be a path
			project.defaultGeodatabase = default_gdb

	project.save()

	# return the project path, setup will continue from here
	return new_temp_project


def make_layer_with_file_symbology(feature_class, layer_file, layer_name=None):
	"""
		Given a feature class or raster and a template layer file with symbology, returns a new Layer object that has the layer
		from the layer file with the feature class as its data source. Optionally, layer can be renamed with layer_name

		This function will be scheduled for deprecation as it's superceded by using the Layer object with a template layer.
		More testing needs to be done and maybe more code written in ArcMap to make the alternative apply to both cases. This will
		be deprecated when that's complete
	:param feature_class:
	:param layer_file:
	:param layer_name:
	:return:
	"""
	#if PRO:  # we'll actually issue the warning once there's a better cross-platform solution. For now, comment it out
	#	warnings.warn("make_layer_with_file_symbology is deprecated in ArcGIS - use Layer objects with a template layer instead", DeprecationWarning)

	layer = None
	if PRO:
		layer_file = mp.LayerFile(layer_file)
		for layer in layer_file.listLayers():  # gets the first layer in the layer file
			break
	else:
		layer = mapping.Layer(layer_file)

	if layer is None:
		raise LayerNotFoundError("No layer available for copying from layer file")
	elif not layer.supports("DATASOURCE"):
		raise NotSupportedError("Provided layer file doesn't support accessing or setting the data source")

	if PRO:
		layer.dataSource = feature_class
	else:
		desc = arcpy.Describe(feature_class)
		if desc.extension and desc.extension != "":  # get the name with extension for replacing the data source
			name = "{}.{}".format(desc.baseName, desc.extension)
		else:
			name = desc.baseName
		layer.replaceDataSource(desc.path, get_workspace_type(feature_class), name)
	if layer_name:
		layer.name = layer_name

	return layer


def reproject_extent(extent, current_extent):
	"""
		Changes an extent from its current spatial reference to the spatial reference on another extent object
	:param extent:
	:param current_extent:
	:return:
	"""
	return extent.projectAs(current_extent.spatialReference)


def get_workspace_type(dataset_path, factory_type=False):
	"""
		Gives us a workspace type that's usable for layer.replaceDataSource in ArcMap based on a dataset path
	:param dataset_path: path to dataset to return workspace type from
	:param factory_type: boolean flag indicating whether to return the workspace_factory type or the standard dataset type - workspace factory values are not yet fully implemented
	:return:
	"""

	"""
		Full list of possible factory values is the following in Pro, per Craig Williams:
		Access, De-limited Text File, File Geodatabase, OLE Database, Raster, ArcInfo, SDE, Shape File, LASDataset, Sql, TrackingServer, NetCDF, SqlLite, FeatureService, Cad, Excel, Street Map, SDC, Custom, WFS
	"""

	if factory_type:
		attr = "factory"
	else:
		attr = "workspace_type"

	prog_id_mapping = {
		"esriDataSourcesGDB.AccessWorkspaceFactory": {
			"workspace_type": "ACCESS_WORKSPACE",
			"factory": "Access"  # not positive this is what this value is - can confirm by loading a MDB layer and testing layer.connectionProperties
		},
		"esriDataSourcesGDB.FileGDBWorkspaceFactory": {
			"workspace_type": "FILEGDB_WORKSPACE",
			"factory": "File Geodatabase"
		},
		"esriDataSourcesGDB.InMemoryWorkspaceFactory": {
			"workspace_type": "NONE",
			"factory": "",
		},
		"esriDataSourcesGDB.SdeWorkspaceFactory": {
			"workspace_type": "SDE_WORKSPACE",
			"factory": "SDE"
		}
	}

	type_mapping = {
		"SHAPEFILE": {
			"workspace_type": "SHAPEFILE_WORKSPACE",
			"factory": "Shape File"  # found by loading a shapefile layer and testing layer.connectionProperties
		},
		"EXCEL": {
			"workspace_type": "EXCEL_WORKSPACE",
			"factory": "Excel"
		},
		"TEXT": {
			"workspace_type": "TEXT_WORKSPACE",
			"factory": "De-limited Text File"
		},
		"RASTER": {
			"workspace_type": "RASTER_WORKSPACE",
			"factory": "Raster"
		},
		"TIN": {
			"workspace_type": "TIN_WORKSPACE",
			"factory": ""
		}
	}

	dataset_desc = arcpy.Describe(dataset_path)
	workspace = dataset_desc.path
	workspace_desc = arcpy.Describe(workspace)

	if workspace_desc.workspaceFactoryProgID.replace(".1", "") in prog_id_mapping:  # if we have the specific name for it here, return that first
		return prog_id_mapping[workspace_desc.workspaceFactoryProgID][attr]
	elif workspace_desc.workspaceType == "FileSystem":
		if dataset_desc.extension == "shp":
			return type_mapping["SHAPEFILE"][attr]
		elif dataset_desc.extension in ("xls", "xlsx"):
			return type_mapping["EXCEL"][attr]
		elif dataset_desc.extension in ("tab", "csv", "txt"):  # probably not the best way to handle this
			return type_mapping["TEXT"][attr]
		elif dataset_desc.dataType == "Raster":
			return type_mapping["RASTER"][attr]
	elif dataset_desc.dataType == "Tin":
		return type_mapping["TIN"][attr]
	elif workspace_desc.workspaceFactoryProgID == "":
		return type_mapping["SHAPEFILE"][attr]  # if we get to here without returning, it's likely a shapefile - there are a few items missing from this conditional - CAD, VPF, etc


def get_workspace_factory_of_dataset(dataset_path):
	"""
		Provides the workspace factory type of a provided dataset - a convenience function that calls get_workspace type with the appropriate flag in the backgroun
	:param dataset_path: path to dataset to return workspace_factory value of
	:return:
	"""
	return get_workspace_type(dataset_path=dataset_path, factory_type=True)



from .version import __version__

__name__ = "amaptor"

import logging
log = logging.getLogger("amaptor")

from amaptor.version_check import PRO, ARCMAP, MAP_EXTENSION, mapping, mp

from amaptor.errors import *  # I know it's bad practice. These will all be uniquely named (I know...)
from . import functions

from amaptor.classes.map_frame import MapFrame
from amaptor.classes.layout import Layout
from amaptor.classes.project import Project
from amaptor.classes.map import Map
from amaptor.classes.layer import Layer






from setuptools import setup
import os
import sys

ON_RTD = os.environ.get('READTHEDOCS', None) == 'True'
try:
	import arcpy
	include_package_data = True
except ImportError:
	if ON_RTD:  # if we're on read-the-docs, then we need to mock out arcpy, otherwise raise the exception
		class Mock(object):
			def __init__(self, *args, **kwargs):
				pass

			def __call__(self, *args, **kwargs):
				return Mock()

			@classmethod
			def __getattr__(cls, name):
				if name in ('__file__', '__path__'):
					return '/dev/null'
				elif name[0] == name[0].upper():
					mockType = type(name, (), {})
					mockType.__module__ = __name__
					return mockType
				else:
					return Mock()


		MOCK_MODULES = ['arcpy']
		for mod_name in MOCK_MODULES:
			sys.modules[mod_name] = Mock()

		include_package_data = False
	else:
		raise

try:
	from amaptor.version import __version__, __author__
except RuntimeError:  # added so it can be installed with setup.py develop when signed in as an admin that's not signed into Portal in ArcGIS Pro
	__version__ = "0.1.1.2"
	__author__ = "nickrsan"

setup(name="amaptor",
	version=__version__,
	description="A compatibility layer (or adaptor) for mapping functions in ArcGIS 10.x (arcpy.mapping/Python 2) and"
				" ArcGIS Pro (arcpy.mp/Python 3)",
	long_description="""This library is an abstraction layer, with its own classes, to access mapping functions, regardless
	of which version of ArcGIS you are using. Think of it as the "six" package, but for arcpy.mapping/arcpy.mp. If you
	write your mapping code against this pacakge, it will work no matter which version of ArcGIS your end users are
	running. In general, the API adheres closely to the ArcGIS Pro api, since that has a cleaner, object-oriented design,
	but it may included differences. Further, methods are lowercased and underscored instead of camelcased, partially to
	make it easy to see at a glance that it's not the same code, and partially due to author preference.

	Documentation can be found at http://amaptor.readthedocs.io
	""",
	packages=['amaptor', 'amaptor.classes', ],
	requires=["arcpy"],
	author=__author__,
	author_email="nrsantos@ucdavis.edu",
	url='https://github.com/ucd-cws/amaptor',
	include_package_data=include_package_data,
)



import logging
log = logging.getLogger("amaptor")

import arcpy

from amaptor.version_check import PRO, ARCMAP, mapping, mp
from amaptor.errors import NotSupportedError, EmptyFieldError, LayerNotFoundError
from amaptor.functions import get_workspace_type, get_workspace_factory_of_dataset
from amaptor.constants import _BLANK_FEATURE_LAYER, _BLANK_RASTER_LAYER

class Layer(object):
	"""
		This object corresponds to arcpy Layers - it theoretically supports the full range of API calls for Layer objects
		but with a major caveat that only some of that has been specifically written for amaptor. The remaining calls
		get passed straight through to the underlying Layer object, and this behavior is subjec to change as more of the
		object is officially supported. When using amaptor Layers (and the rest of amaptor) take note of the version
		you are using so that if the API changes (it will), you can continue to run your code. We'll try to make sensible
		evolutions that help with things and harm as little prior code as possible.

		This object is new and not as well-tested. Existing amaptor functions should now return amaptor.Layer objects,
		but the ability to work with either amaptor layers or ArcGIS native layers is preserved in many cases throughout
		code, both for backwards compatibility and for future convenience, where you might want to
	"""
	def __init__(self, layer_object_or_file, name=None, map_object=None, template_layer=None):
		"""
			Create a Layer object by providing an ArcGIS layer instance, an ArcGIS layer file, or a data source.

		:param layer_object_or_file: an actual instance of a layer object, a valid data source (feature class, raster, etc) or a layer file path (layer file paths work best in Pro, which supports multiple layers in a single file - for cross platform usage, open the layer file and get the Layer object you need, then make an amaptor layer with that)
		:param name: used when loading from a file in Pro to select the layer of interest
		:param map_object: the map this layer belongs to - optional but used when updating symbology in ArcMap - not necessary if you plan to use map.add_layer or map.insert_layer before updating symbology
		:param template_layer: This is used in Pro when constructing a layer from a data source - it will start automatically
					with this layer's properties, symbology, etc. In future versions, we hope to have it autodetect the most appropriate
					template layer that comes with amaptor, but for now, this is an option so that you can get the right properties
					immediately.
		"""
		self.init = False  # we'll set to True when done with init - provides a flag when creating a new layer from scratch in Pro, that we're loading a blank layer
		self.layer_object = None

		if PRO and isinstance(layer_object_or_file, arcpy._mp.Layer):
			self.layer_object = layer_object_or_file
		elif ARCMAP and isinstance(layer_object_or_file, mapping.Layer):
			self.layer_object = layer_object_or_file
		elif PRO:  # otherwise, assume it's a path and run the import for each.
			if layer_object_or_file.endswith(".lyr") or layer_object_or_file.endswith(".lyrx"):
				layer_file = mp.LayerFile(layer_object_or_file)
				for layer in layer_file.listLayers():  # gets the specified layer from the layer file OR the last one
					self.layer_object = layer
					if name and layer.name == name:
						break
			else:  # handle the case of providing a data source of some sort - TODO: Needs to do more checking and raise appropriate exceptions (instead of raising ArcGIS' exceptions)
				# In Pro this is complicated - we can't initialize Layers directly, so we'll use a template for the appropriate data type, then modify it with our information
				desc = arcpy.Describe(layer_object_or_file)
				if not template_layer:
					if desc.dataType in ("FeatureClass", "ShapeFile"):
						layer_file = _BLANK_FEATURE_LAYER
					elif desc.dataType in ("RasterDataset", "RasterBand"):
						layer_file = _BLANK_RASTER_LAYER
					else:
						raise NotSupportedError(
							"This type of dataset isn't supported for initialization in amaptor via ArcGIS Pro")
				else:
					layer_file = template_layer

					avail_layer = arcpy.mp.LayerFile(layer_file)
					arcgis_template_layer = avail_layer.listLayers()[0]

					if arcgis_template_layer is None:
						raise LayerNotFoundError("No layer available for copying from layer file")
					elif not arcgis_template_layer.supports("DATASOURCE"):
						raise NotSupportedError("Provided layer file doesn't support accessing or setting the data source")

				self.layer_object = arcgis_template_layer  # set the layer object to the template
				self._set_data_source(layer_object_or_file)  # now set the data source to be the actual source data - self.data_source does the annoying magic behind this in Pro
				self.name = desc.name  # set the name to the dataset name, as would be typical - just a simple default
				del desc
		else:
			self.layer_object = mapping.Layer(layer_object_or_file)

		self.map = map_object

	@property
	def name(self):
		return self.layer_object.name

	@name.setter
	def name(self, value):
		self.layer_object.name = value

	@property
	def data_source(self):

		if not self.layer_object.supports("DATASOURCE"):
			raise NotSupportedError("Provided layer doesn't support accessing or setting the data source")

		return self.layer_object.dataSource

	@data_source.setter
	def data_source(self, new_source):
		self._set_data_source(new_source)

	def _set_data_source(self, new_source):
		if not self.layer_object.supports("DATASOURCE"):
			raise NotSupportedError("Provided layer file doesn't support accessing or setting the data source")

		desc = arcpy.Describe(new_source)
		if desc.extension and desc.extension != "":  # get the name with extension for replacing the data source
			name = "{}.{}".format(desc.baseName, desc.extension)
		else:
			name = desc.baseName

		if PRO:
			old_connection_properties = self.layer_object.connectionProperties

			new_factory_type = get_workspace_factory_of_dataset(new_source)
			self.layer_object.updateConnectionProperties(
				old_connection_properties,
				{
					'dataset': desc.name,
					'connection_info': {'database': desc.path},
					'workspace_factory': new_factory_type
				}
			)
		else:
			self.layer_object.replaceDataSource(desc.path, get_workspace_type(new_source), name)

	@property
	def symbology(self):
		return self.layer_object.symbology

	@symbology.setter
	def symbology(self, symbology):
		"""
			Updates layer symbology based on copying from:
			 1) amaptor Layer object
			 2) arcpy.mapping/arcpy.mp Layer objects
			 3) symbology object. Symbology objects only exist in Pro, so take care when using them for cross-platform support.

			raises NotSupportedError if the provided object cannot be copied from. If you wish to copy symbology from a Layer
			file, open it and retrieve the appropriate Layer object and pass it here.

			IMPORTANT: If you are setting symbology using this method in ArcMap, you MUST attach an amaptor.Map instance
			representing the Data Frame that this layer is within as this_layer.map *before* doing any symbology operations.
			amaptor functions handle this by default when finding layers and inserting them,  but if you are creating
			your own amaptor layer objects and haven't yet inserted it into a map, you'll need to set the `map` attribute

			```
			  	my_layer = amaptor.Layer("my_layer_name", template_layer="some_layer_file.lyr")
			  	my_layer.map = my_map  # set the map attribute so it knows what data frame to use. Should be an amaptor.Map object, not an actual data frame.
			  	my_layer.symbology = symbol_layer # copies symbology from symbol_layer to my_layer
			```
			The step `my_layer.map` isn't necessary in the instance that you use map.add_layer or map.insert_layer before updating symbology


		:param symbology: Symbology can be a symbology object or a layer to copy it from
		:return:
		"""
		if PRO:
			if isinstance(symbology, arcpy._mp.Symbology):
				new_symbology = symbology
			elif isinstance(symbology, arcpy._mp.Layer) or isinstance(symbology, Layer):  # if it's an amaptor layer, and we're Pro, copy it from there
				new_symbology = symbology.symbology
			else:
				raise NotSupportedError("Cannot retrieve symbology from the object provided. Accepted types are amaptor.Layer, arcpy.mp.Symbology, and arcpy.mp.Layer. You provided {}".format(type(symbology)))
			self.layer_object.symbology = new_symbology
			#self.layer_object.symbology.updateRenderer(new_symbology.renderer.type)  # only used in 2.0+
			#self.layer_object.symbology.updateColorizer(new_symbology.colorizer.type)
		else:  # if ArcMap, we need to do some workaround

			from amaptor.classes.map import Map  # if we put this at the top, we get a circular import - need it to run at runtime for checking - this should be refactored, but not immediately sure how since these classes are mostly appropriately isolated, but bidrectionally reference each other
			if self.map is None or not isinstance(self.map, Map):
				raise EmptyFieldError("map", "Layer is not attached to an amaptor.Map instance - cannot change symbology. See documentation.")

			if isinstance(symbology, Layer):
				source_data = symbology.layer_object
			elif isinstance(symbology, arcpy.mapping.Layer):
				source_data = symbology
			else:
				raise NotSupportedError("Cannot retrieve symbology from the object provided. Accepted types are amaptor.Layer and arcpy.mapping.Layer. You provided {}".format(type(symbology)))

			arcpy.mapping.UpdateLayer(data_frame=self.map.map_object,
									  update_layer=self.layer_object,
									  source_layer=source_data,
									  symbology_only=True)

	def __getter__(self, key):
		"""
			Helps this to be a standin where layers were used before because it will behave as expected for attributes
			of arcpy layers.
		:return:
		"""
		return getattr(self.layer_object, key)

	def __setter__(self, key, value):
		"""
			Helps this to be a standin where layers were used before because it will behave as expected for attributes
			of arcpy layers.
		:return:
		"""
		return setattr(self.layer_object, key, value)


import os
import logging
log = logging.getLogger("amaptor")

import arcpy

from amaptor.version_check import PRO, mapping, mp
from amaptor.errors import *
from amaptor.functions import make_layer_with_file_symbology, reproject_extent
from amaptor.classes.map_frame import MapFrame
from amaptor.classes.layout import Layout
from amaptor.classes.layer import Layer

class Map(object):
	"""
		Corresponds to an ArcMap Data Frame or an ArcGIS Pro Map
	"""
	def __init__(self, project, map_object):

		self.map_object = map_object
		self.project = project
		self.layers = []

		self.list_layers()

		self.frames = []
		self.layouts = []

		self._index_frames()

	@property
	def name(self):
		return self.map_object.name

	@name.setter
	def name(self, value):
		self.map_object.name = value

	def _index_frames(self):
		self.frames = []
		self.layouts = []
		for layout in self.project.layouts:
			for frame in layout.frames:
				if frame.map and frame.map.name == self.name:  # if the frame has a map and it has the same name
					self.frames.append(frame)
					if layout not in self.layouts:
						self.layouts.append(layout)


	def _get_layers_pro(self):
		self._arcgis_layers = self.map_object.listLayers()
		self.layers = [Layer(layer) for layer in self._arcgis_layers]
		for layer in self.layers:  # set the map on the layer as a backreference
			layer.map = self

	def _get_layers_arcmap(self):
		self._arcgis_layers = mapping.ListLayers(self.project.map_document)
		self.layers = [Layer(layer) for layer in self._arcgis_layers]
		for layer in self.layers:  # set the map on the layer as a backreference
			layer.map = self

	def list_layers(self):
		"""
			Returns the list of layers in the map or data frame. Also available as map.layers
		:return:
		"""
		if PRO:
			self._get_layers_pro()
		else:
			self._get_layers_arcmap()

		return self.layers

	def add_layer(self, add_layer, add_position="AUTO_ARRANGE"):
		"""
			Straight replication of addLayer API in arcpy.mp and arcpy.mapping. Adds a layer to a specified position
			in the table of contents.
		:param add_layer: The layer to add to the map
		:param add_position: The position to add the layer into. The default is AUTO_ARRANGE, and available options are the same
			as those available on addLayer.
		:return: None
		"""
		if isinstance(add_layer, Layer):  # if it's an amaptor layer
			new_layer = add_layer.layer_object
			new_layer.map = self  # this helps some future operations that require knowing what map the layer is a part of
		else:
			new_layer = add_layer

		if PRO:
			self.map_object.addLayer(new_layer, add_position)
		else:
			arcpy.mapping.AddLayer(self.map_object, new_layer, add_position)

		self.list_layers()  # make sure the internal layer list is up to date


	def insert_layer(self, reference_layer, insert_layer_or_layerfile, insert_position="BEFORE"):
		"""
			Inserts a layer to a specific position in the table of contents, based on a reference layer.
		:param reference_layer: The arcpy Layer instance to use as the reference layer
		:param insert_layer_or_layerfile: The arcpy Layer instance to insert
		:param insert_position: the position relative to the reference layer to insert the new layer. Default is "BEFORE" (above).
		 	options correspond to those available on insertLayer in arcpy.mapping and arcpy.mp
		:return: None
		"""

		if isinstance(insert_layer_or_layerfile, Layer):
			insert_layer_or_layerfile = insert_layer_or_layerfile.layer_object
			insert_layer_or_layerfile.map = self  # this helps some future operations where layers need to know what maps they're a part of
		else:
			if PRO:
				layer_type = arcpy._mp.Layer
			else:
				layer_type = arcpy._mapping.Layer

			if not isinstance(insert_layer_or_layerfile, layer_type):
				raise(RuntimeError, "provided object is not an Layer instance and can't be added to a map")

		if isinstance(reference_layer, Layer):
			reference_layer = reference_layer.layer_object

		if PRO:
			self.map_object.insertLayer(reference_layer, insert_layer_or_layerfile=insert_layer_or_layerfile, insert_position=insert_position)
		else:
			mapping.InsertLayer(self.map_object, reference_layer, insert_layer_or_layerfile, insert_position)

		# update the internal layers list at the end
		self.list_layers()

	def set_extent(self, extent_object, set_frame="ALL", add_buffer=True, buffer_factor=.05):
		"""
			Sets map frames to a provided extent object. In ArcMap, just sets the data frame's extent. In Pro, it has many
			potential behaviors. If set_frame == "ALL" it sets all map frames linked to this map to this extent (default
			behavior) and sets the default camera for this map so that future map frames will use the same extent.
			If set_frame is an arcpy.mp MapFrame object instance, then it only sets the extent on that map frame.
		:param extent_object: an arcpy.Extent object. It will be reprojected to the spatial reference of the map frame or data frame automatically.
		:param set_frame: ignored in arcmap, behavior described in main method description.
		:param add_buffer: adds an empty space of 5% of the distance across the feature class around the provided extent
		:param buffer_factor: if add_buffer is True, then this factor controls how much space to add around the layer (default=.05)
		:return: None
		"""

		if add_buffer:
			x_buf = (extent_object.XMax - extent_object.XMin) * buffer_factor
			y_buf = (extent_object.YMax - extent_object.YMin) * buffer_factor

			extent_object.XMax += x_buf
			extent_object.XMin -= x_buf
			extent_object.YMax += y_buf
			extent_object.YMin -= y_buf

		if PRO:
			if set_frame == "ALL":
				for frame in self.frames:
					extent = reproject_extent(extent_object, frame.get_extent())
					frame.set_extent(extent)
					self.map_object.defaultCamera.setExtent(extent)
			else:
				if isinstance(set_frame, MapFrame):
					set_frame = set_frame._map_frame_object

				if isinstance(set_frame, arcpy._mp.MapFrame):
					extent = reproject_extent(extent_object, set_frame.camera.getExtent())
					set_frame.camera.setExtent(extent)
				else:
					raise ValueError("Invalid parameter set_frame. It can either be \"ALL\" or an instance of an arcpy.mp MapFrame object")
		else:
			self.map_object.extent = reproject_extent(extent_object, self.map_object.extent)

	def zoom_to_layer(self, layer, set_frame="ALL", add_buffer=True, buffer_factor=.05):
		"""
			Given a name of a layer as a string or a layer object, zooms the map extent to that layer
			WARNING: In Pro, see the parameter information for set_layout on the set_extent method for a description
			of how this option behaves. Since maps don't correspond 1:1 to layouts, in some cases multiple layouts will
			be changed.
		:param layer: can be a string name of a layer, or a layer object
		:param set_layout: PRO ONLY, but ignored in ArcMap, so can be safe to use. This parameter controls which map frames
			are changed by the Zoom to Layer. By default, all linked map frames are updated. If an arcpy.mp.MapFrame instance
			or an amaptor.MapFrame instance is provided, it zooms only that map frame to the layer.
		:param add_buffer: adds an empty space of 5% of the distance across the feature class around the provided extent
		:param buffer_factor: if add_buffer is True, then this factor controls how much space to add around the layer (default=.05)
		:return: None
		"""
		if isinstance(layer, Layer):
			layer = layer.layer_object  # get the actual layer object for the rest of this if an amaptor Layer is passed in.

		if PRO:
			if not isinstance(layer, arcpy._mp.Layer):
				layer = self.find_layer(name=layer).layer_object
			self.set_extent(arcpy.Describe(layer.dataSource).extent, set_frame=set_frame, add_buffer=add_buffer, buffer_factor=buffer_factor)
		else:
			if not isinstance(layer, arcpy._mapping.Layer):
				layer = self.find_layer(name=layer).layer_object
			self.set_extent(layer.getExtent(), add_buffer=add_buffer, buffer_factor=buffer_factor)
			arcpy.RefreshActiveView()

	def insert_layer_by_name_or_path(self, insert_layer_or_layer_file, near_name=None, near_path=None, insert_position="BEFORE"):
		"""
			Not a standard arcpy.mapping or arcpy.mp function - given a name or data source path of a layer, finds it in the layers, and inserts it.
			Only provide either near_name or near_path. If both are provided, near_path will be used because it's more specifci
			:param insert_layer_or_layer_file:
			:param near_name:
			:param near_path:
			:param insert_position:
		:return: None
		"""

		reference_layer = self.find_layer(name=near_name, path=near_path)
		self.insert_layer(reference_layer=reference_layer, insert_layer_or_layerfile=insert_layer_or_layer_file, insert_position=insert_position)

	def insert_feature_class_with_symbology(self, feature_class, layer_file, layer_name=None, near_name=None, near_path=None, insert_position="BEFORE"):
		"""
			Given a path to a feature calss, and a path to a layer file, creates a layer with layer file symbology and
			inserts it using insert_layer_by_name_or_path's approach
		:param feature_class:
		:param layer_file:
		:param layer_name:
		:param near_name:
		:param near_path:
		:param insert_position:
		:return:
		"""

		layer = make_layer_with_file_symbology(feature_class=feature_class, layer_file=layer_file, layer_name=layer_name)
		self.insert_layer_by_name_or_path(layer, near_name=near_name, near_path=near_path, insert_position=insert_position)

	def find_layer(self, name=None, path=None, find_all=False):
		"""
			Given the name OR Path of a layer in the map, returns the layer object. If both are provided, returns based on path.
			If multiple layers with the same name/path exist, returns the first one, unless find_all is True - then it returns a list with all instances.
			Automatically converted to work with new Layer object because self.layers uses them
		:param name:
		:param path:
		:param find_all:
		:return: arcpy.Layer object
		"""
		layers = []
		for layer in self.layers:
			if path is not None and layer.supports("DATASOURCE") and layer.dataSource == path:
				if find_all:
					layers.append(layer)
				else:
					return layer
			elif name is not None and layer.name == name:
				if find_all:
					layers.append(layer)
				else:
					return layer

		if len(layers) == 0:  # basically, we should only be here if find_all is True and find_all came up empty
			raise LayerNotFoundError("Layer with provided name {} or path {} not found".format(name, path))

		return layers

	def _export(self, out_path, layout, mapping_function, mp_function, layout_function, extension, **kwargs):
		"""
			Defines general export behavior for most map export types. Designed to be called only by other methods
			on this class, which will define the functions needed for this.

			For ArcMap documents, behavior is straightforward. An export of the given type is made to out_path.

			In Pro, we have a few options. If layout is an ArcGIS Pro Layout instance or an amaptor Layout instance, then
			only that layout is exported to the output path. If layout is the keyword "ALL" then all layouts associated
			with this map are exported to the output path, with the layout name appended to the path before the extension.

		:param out_path: The path to export the document to
		:param layout: PRO only, safely ignored in ArcMap. The mp.Layout or amaptor.Layout object to export, or the keyword "ALL"
		:param mapping_function: the function to use for export in arcpy.mapping
		:param mp_function: the function to use for export for an arcpy.mp.Layout instance
		:param layout_function: the function to use for export for an amaptor.Layout instance
		:param extension: the file extension to assign the export, without a "." included
		:param kwargs: kwargs that get passed straight through to the exporting functions.
		:return: list of file paths. In most cases, the list will have only one item, but in the case of layout="ALL",
			the list will have many paths generated by the export.
		"""
		file_paths = []
		if ARCMAP:
			function = getattr(mapping, mapping_function)
			function(self.project.map_document, out_path, **kwargs)
			file_paths.append(out_path)
		else:
			if isinstance(layout, arcpy._mp.Layout) or isinstance(layout, Layout):
				if isinstance(layout, Layout):
					layout_function = getattr(layout, layout_function)
					layout_function(layout, out_path, **kwargs)
				else:
					mp_function = getattr(layout, mp_function)
					mp_function(layout, out_path, **kwargs)
				file_paths.append(out_path)
			elif layout == "ALL":
				base_path, file_name = os.path.split(out_path)
				file_base = os.path.splitext(file_name)[0]

				layout_function = getattr(Layout, layout_function)
				for layout in self.layouts:
					output_path = os.path.join(base_path, "{}_{}.{}".format(file_base, layout.name, extension))
					layout_function(layout, output_path, **kwargs)
					file_paths.append(output_path)

		return file_paths

	def export_png(self, out_path, resolution=300, layout="ALL", ):
		"""
			See documentation for _export for description of behavior in each version. The specific option here is only
			the resolution to export at.
		:param out_path: The full path to export the document to. Will be modified in the case of layout="ALL". New generated
			paths will be returned by the function as a list.
		:param resolution: the resolution to export the map at
		:param layout:  PRO only, safely ignored in ArcMap. The mp.Layout or amaptor.Layout object to export, or the keyword "ALL"
		:return:
		"""
		return self._export(out_path, layout=layout, mapping_function="ExportToPNG", mp_function="exportToPNG", layout_function="export_to_png", extension="png", resolution=resolution,)

	def export_pdf(self, out_path, layout="ALL", **kwargs):
		"""
			See documentation for _export for description of behavior in each version. kwargs that apply to exporting to PDF
			in ArcMap and ArcGIS Pro apply here.
		:param out_path: The full path to export the document to. Will be modified in the case of layout="ALL". New generated
			paths will be returned by the function as a list.
		:param layout:  PRO only, safely ignored in ArcMap. The mp.Layout or amaptor.Layout object to export, or the keyword "ALL"
		:param **kwargs: accepts the set of parameters that works for both arcmap and arcgis pro. resolution, image_quality,
			image_compression, embed_fonts, layers_attributes, georef_info, jpeg_compression_quality. In the future,
			this may be reengineered to translate parameters with common goals but different names
		:return:
		"""
		new_kwargs = {}
		allowed_kwargs = ["resolution", "image_quality", "image_compression", "embed_fonts", "layers_attributes", "georef_info", "jpeg_compression_quality"]
		for kwarg in allowed_kwargs:
			if kwarg in kwargs.keys():
				new_kwargs[kwarg] = kwargs[kwarg]  # assign any keys in the kwargs that are valid to a new kwarg dict, tossing out others.

		return self._export(out_path, layout=layout, mapping_function="ExportToPDF", mp_function="exportToPDF", layout_function="export_to_pdf", extension="pdf", **new_kwargs)

	def to_package(self, output_file, **kwargs):
		"""
			Though it's not normally a mapping method, packaging concepts need translation between the two versions, so
			we've included to_package for maps and projects. In ArcGIS Pro, project.to_package will create a Project Package
			and map.to_package will create a map package. In ArcMap, both will create a map package. Calling to_package on the map
			will pass through all kwargs to map packaging because the signatures are the same between ArcMap and ArcGIS Pro.
			Sending kwargs to project.to_package will only send to project package since they differ.

		:param output_file:
		:param kwargs:
		:return:
		"""

		log.warning("Warning: Saving map to export package")
		self.project.save()

		if PRO:
			arcpy.PackageMap_management(self.map_object, output_file, **kwargs)
		else:
			arcpy.PackageMap_management(self.project.path, output_file, **kwargs)


	def replace_text(self, text, replacement):
		"""
			Given a string and a replacement value, finds all instances of that text in text elements and titles, and
			replaces those instances with the new value. Useful for creating your own variables like {species} or
			{field_id} in map templates.

			Similar to the project-level replace_text, but behaves slightly differently. In ArcMap, replaces all
			occurrences in the current map document. In Pro, searches all layouts linked to the current layout and
			replaces the string in any text element in those layouts.
		:param text:
		:param replacement:
		:return:
		"""

		if ARCMAP:
			for elm in arcpy.mapping.ListLayoutElements(self.project.primary_document, "TEXT_ELEMENT"):
				elm.text = elm.text.replace(text, replacement)
		else:
			for layout in self.layouts:  # in pro, iterate through Layout objects instead and replace in all
				layout.replace_text(text, replacement)
